package it.inaf.oats.ia2.tapschemamanager.webapp;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.flipkart.zjsonpatch.JsonDiff;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import it.inaf.oats.ia2.tapschemamanager.businesslayer.EntityWrapper;
import it.inaf.oats.ia2.tapschemamanager.businesslayer.EntityWrapperContainer;
import it.inaf.oats.ia2.tapschemamanager.businesslayer.Status;
import it.inaf.oats.ia2.tapschemamanager.businesslayer.TapSchema;
import java.io.IOException;
import java.io.Serializable;
import javax.ws.rs.Consumes;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import javax.enterprise.context.ConversationScoped;
import javax.inject.Inject;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.QueryParam;

/**
 * REST Web Service
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
@Path("tapschema")
@ConversationScoped
public class TapSchemaResource implements Serializable {

    @Inject
    TapSchemaEditingBean tapSchemaEditingBean;
        
    // containers/entity names
    private static final String TAP_SCHEMA = "tapSchema";
    private static final String SCHEMA = "schema";
    private static final String TABLE = "table";
    private static final String COLUMN = "column";

    private static transient final Gson STANDARD_GSON = new GsonBuilder().create();
    private static transient final Gson NULLABLE_GSON = new GsonBuilder().serializeNulls().create();
    private static transient final ObjectMapper mapper = new ObjectMapper();

    private TapSchema getTapSchema() {
        return tapSchemaEditingBean.getTapSchema();
    }

    private String getPageModelPatches() throws IOException {
        EditTapSchemaPageModel newModel = new EditTapSchemaPageModel(getTapSchema());

        JsonNode jn1 = mapper.readTree(NULLABLE_GSON.toJson(tapSchemaEditingBean.getPreviousModel()));
        JsonNode jn2 = mapper.readTree(NULLABLE_GSON.toJson(newModel));

        tapSchemaEditingBean.setPreviousModel(newModel);

        JsonNode patches = JsonDiff.asJson(jn1, jn2);

        return patches.toString();
    }

    private EntityWrapperContainer getEntitiesContainer(String name) {
        TapSchema tapSchema = getTapSchema();
        if (name.equals(TAP_SCHEMA)) {
            return tapSchema;
        } else if (name.equals(SCHEMA)) {
            return tapSchema.getSelectedEntity();
        } else if (name.equals(TABLE)) {
            return tapSchema.getSelectedTable();
        }
        return null;
    }

    private EntityWrapper getEntityWrapper(String name) {
        TapSchema tapSchema = getTapSchema();
        if (name.equals(SCHEMA)) {
            return tapSchema.getSelectedEntity();
        } else if (name.equals(TABLE)) {
            return tapSchema.getSelectedTable();
        } else if (name.equals(COLUMN)) {
            return tapSchema.getSelectedColumn();
        }
        return null;
    }

    @GET
    @Path("model")
    @Produces(MediaType.APPLICATION_JSON)
    public String getPageModelJson() {
        EditTapSchemaPageModel pageModel = new EditTapSchemaPageModel(getTapSchema());
        tapSchemaEditingBean.setPreviousModel(pageModel);
        return NULLABLE_GSON.toJson(pageModel);
    }

    private void selectEntity(String containerName, String entityName) {
        getEntitiesContainer(containerName).selectEntity(entityName);
    }

    @GET
    @Path("select")
    @Produces(MediaType.APPLICATION_JSON)
    public String select(@QueryParam("containerName") String containerName, @QueryParam("entityName") String entityName) throws IOException {
        selectEntity(containerName, entityName);
        return getPageModelPatches();
    }

    class RemoveColumnResult {

        EditTapSchemaPageModel pageModelUpdates;
        boolean toRemove;
    }

    @GET
    @Path("remove")
    @Produces(MediaType.APPLICATION_JSON)
    public String remove(@QueryParam("containerName") String containerName, @QueryParam("entityName") String entityName) throws IOException {
        EntityWrapperContainer container = getEntitiesContainer(containerName);
        container.removeEntityWrapper(entityName);
        return getPageModelPatches();
    }

    class Addable {

        boolean selected;
        String name;

        Addable(String name) {
            this.name = name;
        }
    }

    @GET
    @Path("addables")
    @Produces(MediaType.APPLICATION_JSON)
    public String getAddables(@QueryParam("containerName") String containerName) {
        ArrayList<Addable> addables = new ArrayList<Addable>();
        for (Object name : getEntitiesContainer(containerName).getAddables()) {
            addables.add(new Addable((String) name));
        }
        return STANDARD_GSON.toJson(addables);
    }

    class AddParams {

        String containerName;
        List<String> entities;
    }

    @POST
    @Path("update")
    @Produces(MediaType.APPLICATION_JSON)
    public String update() throws SQLException, IOException {
        TapSchema tapSchema = getTapSchema();
        tapSchema.save();
        return getPageModelPatches();
    }

    @POST
    @Path("add")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public String add(String params) throws SQLException, IOException {
        final AddParams addParams = STANDARD_GSON.fromJson(params, AddParams.class);

        for (String entityName : addParams.entities) {
            getEntitiesContainer(addParams.containerName).addEntityWrapper(entityName);
        }

        return getPageModelPatches();
    }

    class EditParams {

        String entity;
        String key;
        String value;
    }

    @POST
    @Path("edit")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    public String edit(String params) throws SQLException {
        EditParams editParams = STANDARD_GSON.fromJson(params, EditParams.class);
        EntityWrapper entityWrapper = getEntityWrapper(editParams.entity);

        entityWrapper.setValue(editParams.key, editParams.value);
        boolean isChanged = entityWrapper.isChanged(editParams.key);

        // Update previous model
        tapSchemaEditingBean.getPreviousModel().updateEditableField(editParams.entity, editParams.key, editParams.value, isChanged);

        return isChanged + "";
    }

    @GET
    @Path("undoremove")
    @Produces(MediaType.APPLICATION_JSON)
    public String undoremove(@QueryParam("containerName") String containerName) throws SQLException, IOException {
        getEntitiesContainer(containerName).getSelectedEntity().setStatus(Status.ADDED_PERSISTED);
        return getPageModelPatches();
    }

    @GET
    @Path("searchUCD")
    @Produces(MediaType.APPLICATION_JSON)
    public String searchUCD(@QueryParam("description") String description) throws Exception {
        SearchUCDDialog dialog = tapSchemaEditingBean.getSearchUCDDialog();
        if (description == null || description.equals("")) {
            dialog.setDefault();
        } else {
            dialog.search(description);
        }

        return NULLABLE_GSON.toJson(dialog);
    }

    @GET
    @Path("selectUCD")
    @Produces(MediaType.TEXT_PLAIN)
    public String selectUCD(@QueryParam("index") String indexStr) {
        SearchUCDDialog dialog = tapSchemaEditingBean.getSearchUCDDialog();

        if (indexStr == null || indexStr.equals("")) {
            dialog.selectUCD();
        } else {
            dialog.selectUCD(Integer.parseInt(indexStr));
        }

        return dialog.getSelectedUCD();
    }

    @GET
    @Path("explainUCD")
    @Produces(MediaType.TEXT_PLAIN)
    public String explainUCD(@QueryParam("index") String indexStr) throws Exception {
        SearchUCDDialog dialog = tapSchemaEditingBean.getSearchUCDDialog();

        if (indexStr == null || indexStr.equals("")) {
            return dialog.explain();
        }

        return dialog.explain(Integer.parseInt(indexStr));
    }
}
